/*
  THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT
  LAWS OF THE UNITED STATES.  FOR LICENSING INFORMATION CONTACT:

  Christian Bischof or Lucas Roh, Mathematics and Computer Science Division,
  Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL 60439, 
  {bischof,roh}@mcs.anl.gov.
*/

/* 
   This is a specialized version of the Jacobian library that supports only the 
   case when there is one (1) independent variable, i.e., all gradient vectors
   are scalars.
   
   WARNING: this version does not support parallelized derivative computation via MPI 
   or dynamic memory allocation (which doesn't make sense since grad size is always 1). 
*/

#include <stdio.h>
#include <stdlib.h>

#if !defined(AD_GRAD_H)
#   define AD_GRAD_H
#   include <string.h>
    
#include "ad_grad_macro_axpys1.h"

extern int ad_grad_size;
extern int ad_total_grad_size;   /*only used when generating MPI programs*/
extern int ad_grad_size_shadow;

#   if defined(__cplusplus)
        extern "C" {
#   endif
    
#define ad_AD_GradFinal() \
    { \
        ad_grad_size = 0; \
    }

#define ad_AD_GradInit(n) \
    { \
        if (n == -1) \
           ad_grad_size = ad_GRAD_MAX; \
        else \
           ad_grad_size = n; \
        ad_grad_size_shadow = 0; \
    }

#define ad_AD_ClearGrad(gz) memset((char*)(DERIV_grad(*gz)), 0, ad_GRAD_MAX*sizeof(double)); 

#define ad_AD_InitGrad(gz) memset((char*)(DERIV_grad(*gz)), 0, ad_GRAD_MAX*sizeof(double));

#define ad_AD_CopyGrad(gz,gx) gz[0] = gx[0];

#   define ad_AD_GetTotalGradSize() ad_grad_size
#   define ad_AD_SetTotalGradSize(x) \
    { \
	if (x > ad_GRAD_MAX) {\
                fprintf (stderr, "ad_AD_SetTotalGradSize() Error: the number of independent variables %d exceeds the max (GRAD_MAX)!\n", x);\
                fflush(stderr);  \
                abort();  \
        }\
        ad_grad_size_shadow = x;\
        ad_grad_size = x; \
    }

#   define ad_AD_IncrementTotalGradSize(x) \
    { \
         if (x + ad_grad_size_shadow > ad_GRAD_MAX) {\
                fprintf (stderr, "ad_AD_IncrementTotalGradSize() Error: the number of independent variables %d exceeds the max (GRAD_MAX)!\n", x + ad_grad_size_shadow);\
                fflush(stderr);  \
                abort();  \
        }\
        ad_grad_size_shadow += x;\
    }

#   define ad_AD_ExtractGrad(a, var) (a) = DERIV_grad(*var)[0]; 

#   define ad_AD_ExtractVal(a, var) (a) = DERIV_val(var); 
    
#   define ad_AD_SetGrad(a, var) DERIV_grad(var)[0] = (a)[0]; 

#   define ad_AD_SetIndepDone() ad_AD_CommitShadowVar()
#   define ad_AD_ResetIndep() ad_AD_ResetShadowVar()
#   define ad_AD_SetIndep(var) \
    { \
        int pOsItIoN = ad_AD_IncrShadowVar(); \
        if (pOsItIoN > ad_GRAD_MAX) {\
            fprintf (stderr, "ad_AD_SetIndep() Error: the number of " \
		     "independent variables %d exceeds the max (GRAD_MAX)!\n",\
		     pOsItIoN); \
            fflush(stderr);  \
            abort();  \
        }\
        ad_AD_ClearGrad(DERIV_grad(var)); \
        DERIV_grad(var)[pOsItIoN] = 1; \
    }
#   define ad_AD_SetIndepArray(vars, size) \
    { \
        int iWiLlNeVeRCoNfLiCt; \
        for (iWiLlNeVeRCoNfLiCt = 0; iWiLlNeVeRCoNfLiCt < size; \
	       iWiLlNeVeRCoNfLiCt++) { \
            ad_AD_ClearGrad(&((vars)[iWiLlNeVeRCoNfLiCt])); \
            DERIV_grad((vars)[iWiLlNeVeRCoNfLiCt])[ad_AD_IncrShadowVar()] = 1; \
        } \
    }

/* Array of colors should be the same length as the vars array */
#   define ad_AD_SetIndepArrayColored(vars, size, colors) \
    { \
        int iWiLlNeVeRCoNfLiCt; \
        for (iWiLlNeVeRCoNfLiCt = 0; iWiLlNeVeRCoNfLiCt < size; \
	       iWiLlNeVeRCoNfLiCt++) { \
            ad_AD_ClearGrad(DERIV_grad((vars)[iWiLlNeVeRCoNfLiCt])); \
            DERIV_grad((vars)[iWiLlNeVeRCoNfLiCt])[ad_grad_size_shadow+colors[iWiLlNeVeRCoNfLiCt]] = 1; \
        } \
    }

/* values array is the same length as vars */
#   define ad_AD_SetIndepVector(vars, size, values) \
    { \
        int iWiLlNeVeRCoNfLiCt; \
        for (iWiLlNeVeRCoNfLiCt = 0; iWiLlNeVeRCoNfLiCt < size; \
	       iWiLlNeVeRCoNfLiCt++) { \
            ad_AD_ClearGrad(DERIV_grad((vars)[iWiLlNeVeRCoNfLiCt])); \
            DERIV_grad((vars)[iWiLlNeVeRCoNfLiCt])[ad_grad_size_shadow] = values[iWiLlNeVeRCoNfLiCt]; \
        } \
        ad_grad_size_shadow++; \
    }
    
int ad_AD_IncrShadowVar();
void ad_AD_CommitShadowVar();
void ad_AD_ResetShadowVar();

#   if defined(__cplusplus)
        }
#  endif
#endif /*AD_GRAD_H*/


