/* ad_mempool.c

   a data structure for handing out same
   size chunks of memory to adic from a preallocated
   and dynamically expansive pool

   written by mark fashing for boyanna norris and the
   adic project at argonne national laboratory
*/

#include "ad_mempool.h"

#define max(a,b) (a > b ? a : b)

/* macro to dereference what an element of the free list
   points to */
#define next(a) *((void **)a)

mempool ad_grad_mempool;

void mp_init(struct mempool_struct *mp) {
  void *temp, *end;
  int size;

  /* size of memory chunks must be large enough to
     accomodate void* for the free list */
  size = max(sizeof(POOL_TYPE_X), sizeof(void*));

  /* first element of free list is the first
     space in the pool */
  mp->freelist = (void *)mp->pool;

  /* loop until the next to last element */
  end = mp->freelist + POOL_SIZE - size;
  for(temp = mp->freelist; temp < end;
      temp += size) {
    /* the next element in the free list is
       simply the next element in the array */
    next(temp) = temp + size;
  }
  /* terminate free list with NULL */
  next(temp) = NULL;

  /* no additional mempools yet */
  mp->next = NULL;

  /*
  for(temp = mp->freelist; temp != NULL; temp = *((void **)temp))
    fprintf(stdout, "%p -> %p\n", temp, *((void **)temp));
  */

}

/* creates another memory pool and adds it to the
   free list of mp */
int mp_grow(struct mempool_struct *mp) {
  mempool *temp;
  void *ptr;

  /* can't grow something that isn't there */
  if(mp == NULL) return 0;

  /* find the end of the linked list of mempools */
  for(temp = mp; temp->next != NULL; temp = temp->next);
  /* create a new mempool */
  temp->next = (mempool *)malloc(sizeof(mempool));
  temp = temp->next;
  if(temp == NULL) return 0;  /* no space to grow */

  /* set up free list of new mempool */
  mp_init(temp);

  /* add new free list to the end of the old */
  if(mp->freelist == NULL) mp->freelist = temp->freelist;
  else {
    for(ptr = mp->freelist; next(ptr) != NULL; ptr = next(ptr));
    next(ptr) = temp->freelist;
  }

  return 1;  /* return 1 for success */
}

void *mp_alloc(mempool *mp) {
  void *temp;

  /* no memory left to hand out?  grab some more */
  if(mp->freelist == NULL) mp_grow(mp);

  /* hand out first element in free list */
  temp = mp->freelist;

  /* take element out of free list */
  if(temp != NULL)
    mp->freelist = next(temp);

  /* returns memory if successful, NULL if not */
  return temp;
}

void **mp_allocn(mempool *mp, int n, void **diarray) {
  int i;

  for(i = 0; i < n; i++) {
    /* no memory?  grow the mempool */
    if(mp->freelist == NULL) mp_grow(mp);
    /* hand out first element */
    diarray[i] = mp->freelist;
    /* take element out of free list */
    mp->freelist = next(mp->freelist);
  }

  /* return an array of pointers */
  return diarray;
}

void mp_free(mempool *mp, void *ptr) {
  /* prepend memory to free list */
  next(ptr) = mp->freelist;
  mp->freelist = ptr;
}

void mp_freen(mempool *mp, int n, void **diarray) {
  int i;

  /* prepend an array of pointers to free list
     in reverse order, i.e. put them back the
     way you found them */
  for(i = n-1; i >= 0; i--) {
    next(diarray[i]) = mp->freelist;
    mp->freelist = diarray[i];
  }
}

void mp_dumpfree(FILE *fstream, mempool *mp) {
  void *temp;

  /* traverse the free list printing it's
     contents */
  for(temp = mp->freelist; temp != NULL; temp = next(temp))
    fprintf(fstream, "%p -> %p\n", temp, next(temp));
}

void mp_destroy(mempool *mp) {
  mempool *temp = mp->next, *nextinline;

  /* destroy the free list */
  mp->freelist = NULL;

  /* deallocate appended mempools */
  while (temp != NULL) {
    nextinline = temp->next;
    free(temp);
    temp = nextinline;
  }  
}

