#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "cvode.h"
#include "llnltyps.h"
#include "nvector.h"
#include "llnlmath.h"
#include "sensitivity.h"

#define HALF   RCONST(0.5)  /* real 0.5   */
#define ZERO   RCONST(0.0)  /* real 0.0   */
#define ONE    RCONST(1.0)  /* real 1.0   */

void FDQ(integer N, real t, N_Vector u, N_Vector udot, void *s_data)
{
  integer i, ns, ny;
  real pbar, psave;
  real deltap, deltay, rdeltap, rdeltay, delta, r2delta;
  N_Vector *usub, *udotsub, *ewtsub;
  N_Vector ftemp, ytemp;
  CVodeMem cvmem;
  SensData sdata;
  void *fdata;
  real rtol, uround, rho, mu;
  integer method;
  machEnvType machenv;
  N_Vector temp;
  real normu, normv;

  sdata = (SensData) s_data;
  cvmem = (CVodeMem) sdata->cv_mem;
  machenv = (machEnvType) cvmem->cv_machenv;

  usub = (N_Vector *) u->subptr;
  udotsub = (N_Vector *) udot->subptr;
  ewtsub = (N_Vector *) cvmem->cv_ewt->subptr;

  /* Extract parameters and workspace from sdata */
  ns = sdata->ns;
  ny = sdata->ny;
  fdata = sdata->f_data;
  ftemp = sdata->ftemp;
  ytemp = sdata->ytemp;

  rtol = *(cvmem->cv_reltol);
  uround = cvmem->cv_uround;
  mu = sqrt(MAX(rtol, uround));

  /********************************************************************/
  /* Choose Formula For delta In Differencing Formulas                */
  /* 2nd order (centered) or 2 1st order (forward) differences        */
  /*                                                                  */
  /* mu = sqrt(rtol) or mu = sqrt(uround)                             */
  /********************************************************************/
  /* FDQ for derivatives of scaled sensitivities                      */
  /* Use F(t, y + deltay*w, p + deltap*pbar)                          */
  /*                                                                  */
  /* Method 0: deltay = deltap = MIN(mu, 1/norm(w));        2nd order */
  /* Method 1: deltay = MIN(mu, 1/norm(w)) and deltap = mu; 1st order */
  /* Method 2: deltay = deltap = mu                         2nd order */ 
  /* ================================================================ */
  /* FDQ for derivatives of unscaled sensitivities                    */
  /* Use F(t, y + delta*s, p + delta)                                 */
  /*                                                                  */
  /* Method 3: delta = mu*MAX(pbar, 1/norm(u));             2nd order */
  /* Method 4: delta = mu*MAX(pbar, norm(v));               2nd order */
  /* Method 5: delta = mu*pbar;                             2nd order */ 
  /********************************************************************/
  method = 2;
  /********************************************************************/

  /* Compute y derivatives */
  sdata->f_ptr(ny, t, usub[0], udotsub[0], fdata);
  (sdata->nodef)++;

  for (i=1; i <= ns; i++) {
    psave = sdata->p[i-1];
    pbar = sdata->pbar[i-1];
    
    switch (method) {
    case 0: /* Centered Differences For w' */

      /* Set deltap for perturbing parameter i */
      deltap = mu;
      rdeltap = ONE/deltap;

      /* Determine deltay for perturbing y */
      rho = N_VWrmsNorm(usub[i], ewtsub[i]);
      rho = MAX(rho, ONE/mu);
      deltay = ONE/rho;    
      rdeltay = rho;

      delta = MIN(deltap, deltay);
      r2delta = HALF/delta;

      /* Forward perturb y and parameter i */
      N_VLinearSum(delta, usub[i], ONE, usub[0], ytemp);
      sdata->p[i-1] = psave + delta*pbar;

      sdata->f_ptr(ny, t, ytemp, udotsub[i], fdata);
      (sdata->nodef)++;

      /* Backward perturb y and parameter i */
      N_VLinearSum(-delta, usub[i], ONE, usub[0], ytemp);
      sdata->p[i-1] = psave - delta*pbar;

      sdata->f_ptr(ny, t, ytemp, ftemp, fdata);
      (sdata->nodef)++;

      /* Centered differencing for 2nd order accuracy */
      N_VLinearSum(r2delta, udotsub[i], -r2delta, ftemp, udotsub[i]);
  
      /* Restore original value of parameter i */
      sdata->p[i-1] = psave;
      break;

    case 1: /* Forward Differences For Jw and df/dp Terms */

      /* Set deltap for perturbing parameter i */
      deltap = mu;
      rdeltap = ONE/deltap;

      /* Determine deltay for perturbing y */
      rho = N_VWrmsNorm(usub[i], ewtsub[i]);
      rho = MAX(rho, ONE/mu);
      deltay = ONE/rho;    
      rdeltay = rho;

      /* Perturb parameter i only */
      sdata->p[i-1] = psave + deltap*pbar;

      sdata->f_ptr(ny, t, usub[0], ftemp, fdata);
      (sdata->nodef)++;

      /* Restore original value of parameter i */
      sdata->p[i-1] = psave;

      /* Store pbar * df/dp into udotsub[i] */
      N_VLinearSum(rdeltap, ftemp, -rdeltap, udotsub[0], udotsub[i]);

      /* Perturb y only*/
      N_VLinearSum(deltay, usub[i], ONE, usub[0], ytemp);

      sdata->f_ptr(ny, t, ytemp, ftemp, fdata);
      (sdata->nodef)++;

      /* Store Jw into ytemp */
      N_VLinearSum(rdeltay, ftemp, -rdeltay, udotsub[0], ytemp);

      /*** Compute udotsub[i] = Jw + pbar * df/dp ***/
      N_VLinearSum(ONE, ytemp, ONE, udotsub[i], udotsub[i]);
      break;

    case 2:

      delta = mu;
      r2delta = HALF/delta;

      /* Forward perturb y and parameter i */
      N_VLinearSum(delta, usub[i], ONE, usub[0], ytemp);
      sdata->p[i-1] = psave + delta*pbar;

      sdata->f_ptr(ny, t, ytemp, udotsub[i], fdata);
      (sdata->nodef)++;

      /* Backward perturb y and parameter i */
      N_VLinearSum(-delta, usub[i], ONE, usub[0], ytemp);
      sdata->p[i-1] = psave - delta*pbar;

      sdata->f_ptr(ny, t, ytemp, ftemp, fdata);
      (sdata->nodef)++;

      /* Centered differencing for 2nd order accuracy */
      N_VLinearSum(r2delta, udotsub[i], -r2delta, ftemp, udotsub[i]);
  
      /* Restore original value of parameter i */
      sdata->p[i-1] = psave;
      break;

    case 3:
      
      temp = N_VNew(ny, machenv);
      N_VDiv(ewtsub[0], ewtsub[i], temp);
      normu = RSqrt(N_VDotProd(temp, temp));
      N_VFree(temp);

      delta = mu * MAX(ABS(pbar), ONE/normu);
      r2delta = HALF/delta;

      /* Forward perturb y and parameter i */
      N_VLinearSum(delta, usub[i], ONE, usub[0], ytemp);
      sdata->p[i-1] = psave + delta;

      sdata->f_ptr(ny, t, ytemp, udotsub[i], fdata);
      (sdata->nodef)++;

      /* Backward perturb y and parameter i */
      N_VLinearSum(-delta, usub[i], ONE, usub[0], ytemp);
      sdata->p[i-1] = psave - delta;

      sdata->f_ptr(ny, t, ytemp, ftemp, fdata);
      (sdata->nodef)++;

      /* Centered differencing for 2nd order accuracy */
      N_VLinearSum(r2delta, udotsub[i], -r2delta, ftemp, udotsub[i]);
  
      /* Restore original value of parameter i */
      sdata->p[i-1] = psave;
      break;

    case 4:
      
      temp = N_VNew(ny, machenv);
      N_VDiv(ewtsub[i], ewtsub[0], temp);
      normv = RSqrt(N_VDotProd(temp, temp));
      N_VFree(temp);
      
      delta = mu * MAX(ABS(pbar), normv);
      r2delta = HALF/delta;

      /* Forward perturb y and parameter i */
      N_VLinearSum(delta, usub[i], ONE, usub[0], ytemp);
      sdata->p[i-1] = psave + delta;

      sdata->f_ptr(ny, t, ytemp, udotsub[i], fdata);
      (sdata->nodef)++;

      /* Backward perturb y and parameter i */
      N_VLinearSum(-delta, usub[i], ONE, usub[0], ytemp);
      sdata->p[i-1] = psave - delta;

      sdata->f_ptr(ny, t, ytemp, ftemp, fdata);
      (sdata->nodef)++;

      /* Centered differencing for 2nd order accuracy */
      N_VLinearSum(r2delta, udotsub[i], -r2delta, ftemp, udotsub[i]);
  
      /* Restore original value of parameter i */
      sdata->p[i-1] = psave;
      break;

    case 5:
    
      delta = mu*ABS(pbar);
      r2delta = HALF/delta;

      /* Forward perturb y and parameter i */
      N_VLinearSum(delta, usub[i], ONE, usub[0], ytemp);
      sdata->p[i-1] = psave + delta;

      sdata->f_ptr(ny, t, ytemp, udotsub[i], fdata);
      (sdata->nodef)++;

      /* Backward perturb y and parameter i */
      N_VLinearSum(-delta, usub[i], ONE, usub[0], ytemp);
      sdata->p[i-1] = psave - delta;

      sdata->f_ptr(ny, t, ytemp, ftemp, fdata);
      (sdata->nodef)++;

      /* Centered differencing for 2nd order accuracy */
      N_VLinearSum(r2delta, udotsub[i], -r2delta, ftemp, udotsub[i]);
  
      /* Restore original value of parameter i */
      sdata->p[i-1] = psave;
      break;
    }
  }
}

/* SensSetUp routine. Set up block of sensitivity data */

void *SensSetUp(integer ny, integer ns, void *f_data, void *p, void *pbar,
		void(*f_ptr)(integer, real, N_Vector, N_Vector, void *),
		void *machEnv)
{
  SensData sdata;
  machEnvType machenv;

  sdata = (SensData) malloc(sizeof(*sdata));
  machenv = (machEnvType) machEnv;
  sdata->isens = 0;
  sdata->nodef = 0;
  sdata->ny = ny;
  sdata->ns = ns;
  sdata->f_data = f_data;
  sdata->p = p;
  sdata->pbar = pbar;
  sdata->f_ptr = f_ptr; 
  sdata->ftemp = N_VNew(ny, machenv);
  sdata->ytemp = N_VNew(ny, machenv);
  machenv->s_data = sdata;
  return(sdata);
}

/* SensFree routine. Free the block of sensitivity data */

void SensFree(void *s_data)
{
  SensData sdata;

  sdata = (SensData) s_data;
  N_VFree(sdata->ftemp);
  N_VFree(sdata->ytemp);
  free(sdata);
}

/* SensSetAtol routine. Set absolute error tolerances for sensitivities */

void SensSetAtol(N_Vector atol, integer ns)
{
  N_Vector *atolsub;
  int i;

  atolsub = (N_Vector *) atol->subptr;

  /* Load y error tolerances as sensitivity error tolerances */
  for (i=1; i<=ns; i++)
    N_VScale(ONE, atolsub[0], atolsub[i]);
}
