/******************************************************************
 *                                                                *
 * File          : sensitivity.c                                  *
 * Last Modified : 20 August 1999                                 *
 *----------------------------------------------------------------*
 * This is the implementation file for computing the sensitivity  *
 * of the ODE solution y(t, p) with respect to the parameters     *
 * in p.                                                          *
 *                                                                *
 ******************************************************************/


/************************************************************/
/******************* BEGIN Imports **************************/
/************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "cvode.h"
#include "llnltyps.h"
#include "nvector.h"
#include "llnlmath.h"
#include "sensitivity.h"

/************************************************************/
/******************** END Imports ***************************/
/************************************************************/

/************************************************************/
/************** BEGIN Private Constants *********************/
/************************************************************/

#define ZERO      RCONST(0.0)  /* real 0.0   */
#define HALF      RCONST(0.5)  /* real 0.5   */
#define ONE       RCONST(1.0)  /* real 1.0   */

/***************************************************************/
/***************** BEGIN Error Messages ************************/
/***************************************************************/

/* SensCVodeMalloc Error Messages */

#define SensCVM         "SensCVodeMalloc-- "

#define MSG_BAD_NY       SensCVM "ny=%ld < 1 illegal.\n\n"
#define MSG_BAD_NS       SensCVM "ns=%ld < 0 illegal.\n\n"
#define MSG_P_NULL       SensCVM "p=NULL illegal.\n\n"
#define MSG_PBAR_NULL    SensCVM "pbar=NULL illegal.\n\n"

/***************************************************************/
/****************** END Error Messages *************************/
/***************************************************************/

/************************************************************/
/************** END Private Constants ***********************/
/************************************************************/


/**************************************************************/
/********* BEGIN Private Helper Functions Prototypes **********/
/**************************************************************/

/**************************************************************/
/********** END Private Helper Functions Prototypes ***********/
/**************************************************************/

/**************************************************************/
/**************** BEGIN Readability Constants *****************/
/**************************************************************/

#define CENTERED1  0
#define CENTERED2  1
#define FORWARD1   2
#define FORWARD2   3
#define nfe        (cvmem->cv_nfe)

/**************************************************************/
/***************** END Readability Constants ******************/
/**************************************************************/

/***************************************************************/
/************* BEGIN Sensitivity Implementation ****************/
/***************************************************************/

/*******************************************************************/
/******** BEGIN Private Helper Functions Implementation ************/
/*******************************************************************/

/****************** FDQ ********************************************

 This routine evaluates the right hand side of the ODE system
 y' = f(t, y) and approximates the right hand side of the scaled 
 sensitivity  differential equation w'_i = J*w_i + pbar_i*df/dp_i 
 using finite differences.

**********************************************************************/

void FDQ_orig(integer Ntotal, real t, N_Vector u, N_Vector udot, 
                void *s_data)
{
  integer i, ns, ny, method;
  real pbar, psave;
  real delta, rdelta, r2delta;
  real deltap, deltay, rdeltap, rdeltay, r2deltap, r2deltay;
  real rtol, uround, normw;
  real ratio, rhomax;
  N_Vector *usub, *udotsub, *ewtsub;
  N_Vector ftemp, ytemp, vtemp;
  CVodeMem cvmem;
  SensData sdata;
  void *fdata;

  sdata = (SensData) s_data;
  cvmem = (CVodeMem) sdata->cv_mem;

  usub = (N_Vector *) u->subptr;
  udotsub = (N_Vector *) udot->subptr;
  ewtsub = (N_Vector *) cvmem->cv_ewt->subptr;

  /* Extract parameters and workspace from sdata */
  ns = sdata->ns;
  ny = sdata->ny; 
  fdata = sdata->f_data;
  ftemp = sdata->ftemp;
  ytemp = sdata->ytemp;

  rtol = *(cvmem->cv_reltol);
  uround = cvmem->cv_uround;

  /*****************************************/
  /* Set deltap for perturbing parameter i */
  /*****************************************/
  deltap = sqrt(MAX(rtol, uround));
  rdeltap = ONE/deltap;

  /*******************************************************************
   * Compute y derivative as y' = f(t,y).                            *
   *******************************************************************/
  sdata->f_ptr(ny, t, usub[0], udotsub[0], fdata);
  
  for (i = 1; i <= ns; i++) {

  /*******************************************************************
   * Estimate i-th scaled sensitivity derivative w'_i using          *
   * finite difference formulas.                                     *
   *******************************************************************/

    psave = sdata->p[i-1];
    pbar = sdata->pbar[i-1];

    /*************************************/
    /* Determine deltay for perturbing y */
    /*************************************/
    normw = N_VWrmsNorm(usub[i], ewtsub[i]);
    rdeltay = MAX(normw, rdeltap);
    deltay = ONE/rdeltay;

    /***********************************************************/
    /* Select finite difference formula for w'_i               */
    /***********************************************************/
    ratio = deltay*rdeltap;
    rhomax = sdata->rhomax;

    if ((MAX(ONE/ratio, ratio) <= fabs(rhomax)) || rhomax == ZERO)
      /***************************************************/
      /* Centered or forward difference formula for w'_i */
      /***************************************************/
      method = (rhomax >= ZERO) ? CENTERED1 : FORWARD1; 
    else {
      /*****************************************************************/
      /* Centered or forward differences for each term: Jw, pbar*df/dp */
      /*****************************************************************/
      method = (rhomax > ZERO) ? CENTERED2 : FORWARD2;
    }

    switch (method) {

      /**********************************************/
    case 0: /* Centered difference formula for w'_i */
      /**********************************************/
    
      delta = MIN(deltay, deltap);
      r2delta = HALF/delta;

      /* Forward perturb y and parameter i */
      N_VLinearSum(delta, usub[i], ONE, usub[0], ytemp);
      sdata->p[i-1] = psave + delta*pbar;

      sdata->f_ptr(ny, t, ytemp, udotsub[i], fdata);
      nfe++;

      /* Backward perturb y and parameter i */
      N_VLinearSum(-delta, usub[i], ONE, usub[0], ytemp);
      sdata->p[i-1] = psave - delta*pbar;

      sdata->f_ptr(ny, t, ytemp, ftemp, fdata);
      nfe++;

      /* Centered differencing for 2nd order accuracy */
      N_VLinearSum(r2delta, udotsub[i], -r2delta, ftemp, udotsub[i]);
  
      /* Restore original value of parameter i */
      sdata->p[i-1] = psave;

      break;

      /**************************************************************/
    case 1: /* Centered differences for each term: Jw, pbar*df/dp   */
      /**************************************************************/

      r2deltap = HALF/deltap;
      r2deltay = HALF/deltay;
      
      /* Forward perturb y only */
      N_VLinearSum(deltay, usub[i], ONE, usub[0], ytemp);

      sdata->f_ptr(ny, t, ytemp, udotsub[i], fdata);
      nfe++;

      /* Backward perturb y only */
      N_VLinearSum(-deltay, usub[i], ONE, usub[0], ytemp);

      sdata->f_ptr(ny, t, ytemp, ftemp, fdata);
      nfe++;

      /*********************************/
      /* Store Jw term into udotsub[i] */
      /*********************************/
      N_VLinearSum(r2deltay, udotsub[i], -r2deltay, ftemp, udotsub[i]);

      /* Rename ytemp as vtemp for readability */
      vtemp = ytemp;

      /* Forward perturb parameter i only */
      sdata->p[i-1] = psave + deltap*pbar;
      
      sdata->f_ptr(ny, t, usub[0], vtemp, fdata);
      nfe++;
      
      /* Backward perturb parameter i only */
      sdata->p[i-1] = psave - deltap*pbar;

      sdata->f_ptr(ny, t, usub[0], ftemp, fdata);
      nfe++;

      /*********************************/
      /* Store pbar*df/dp into vtemp   */
      /*********************************/
      N_VLinearSum(r2deltap, vtemp, -r2deltap, ftemp, vtemp);

      /*** Compute udotsub[i] = Jw + pbar*df/dp ***/
      N_VLinearSum(ONE, udotsub[i], ONE, vtemp, udotsub[i]);

      /* Restore original value of parameter i */
      sdata->p[i-1] = psave;

      break;

      /**********************************************/
    case 2: /* Forward difference formula for w'_i  */
      /**********************************************/
    
      delta = MIN(deltay, deltap);
      rdelta = ONE/delta;

      /* Forward perturb y and parameter i */
      N_VLinearSum(delta, usub[i], ONE, usub[0], ytemp);
      sdata->p[i-1] = psave + delta*pbar;

      sdata->f_ptr(ny, t, ytemp, ftemp, fdata);
      nfe++;

      /* Forward differencing for 1st order accuracy */
      N_VLinearSum(rdelta, ftemp, -rdelta, udotsub[0], udotsub[i]);
  
      /* Restore original value of parameter i */
      sdata->p[i-1] = psave;

      break;

      /*************************************************************/
    case 3: /* Forward differences for each term: Jw, pbar*df/dp   */
      /*************************************************************/

      /* Perturb y only*/
      N_VLinearSum(deltay, usub[i], ONE, usub[0], ytemp);

      sdata->f_ptr(ny, t, ytemp, ftemp, fdata);
      nfe++;

      /****************************/
      /* Store Jw into udotsub[i] */
      /****************************/
      N_VLinearSum(rdeltay, ftemp, -rdeltay, udotsub[0], udotsub[i]);

      /* Perturb parameter i only */
      sdata->p[i-1] = psave + deltap*pbar;

      sdata->f_ptr(ny, t, usub[0], ftemp, fdata);
      nfe++;

      /* Restore original value of parameter i */
      sdata->p[i-1] = psave;

      /* Rename ytemp as vtemp for readability */
      vtemp = ytemp;

      /*********************************/
      /* Store pbar*df/dp into vtemp   */
      /*********************************/
      N_VLinearSum(rdeltap, ftemp, -rdeltap, udotsub[0], vtemp);

      /*** Compute udotsub[i] = Jw + pbar*df/dp ***/
      N_VLinearSum(ONE, udotsub[i], ONE, vtemp, udotsub[i]);

      break;

    default:      
      /* do nothing */
      break;
    }
  }
}

/*******************************************************************/
/********* END Private Helper Functions Implementation *************/
/*******************************************************************/


/***************************************************************/
/************** END Sensitivity Implementation *****************/
/***************************************************************/

